/*----------------------------------------------------------------------
  File    : lvqx.c
  Contents: learning vector quantization execution
  Author  : Christian Borgelt
  History : 2003.03.11 file created from file lvqx.c
            2003.05.15 normalization parameters moved to lvq file
            2003.08.16 slight changes in error message output
            2004.02.26 source files mlvqx.c and lvqx.c combined
            2007.01.10 use of attribute maps changed
            2007.02.14 adapted to modified module tabread
            2007.03.16 table and matrix version combined
            2013.08.19 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output format changed to significant digits
            2013.08.29 adapted to new function as_target()
            2013.08.30 missing deallocation of pattern buffer added
            2014.10.24 changed from LGPL license to MIT license
            2018.09.10 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef LVQ_PARSE
#define LVQ_PARSE
#endif
#include "lvq.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

#ifdef _MSC_VER
#ifndef snprintf
#define snprintf _snprintf
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "lvqx"
#define DESCRIPTION "learning vector quantization execution"
#define VERSION     "version 2.2 (2018.10.16)         " \
                    "(c) 2003-2018   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse errors on input file */
#define E_VECSIZE  (-10)        /* invalid pattern size */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction result --- */
  DIMID  cnt;                   /* number of reference vectors */
  DIMID  vec;                   /* predicted reference vector */
  CCHAR *col_vec;               /* name   of vector     column */
  int    cwd_vec;               /* width  of vector     column */
  CCHAR *col_act;               /* name   of activation column(s) */
  int    cwd_act;               /* width  of activation column(s) */
  int    dig_act;               /* digits of activation column(s) */
} RESULT;                       /* (prediction result) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_VECSIZE -10 */  "invalid vector size %"DIMID_FMT,
  /* E_UNKNOWN -11 */  "unknown error",
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner for network description */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static ATTMAP   *attmap = NULL; /* attribute map */
static TABLE    *table  = NULL; /* data table */
static LVQNET   *lvq    = NULL; /* learning vector quantization net */
static RESULT   res     = {     /* prediction result */
  0, 0,                         /* number of reference vectors */
  NULL, 0,                      /* data for vector     column */
  NULL, 0, 6 };                 /* data for activation column(s) */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (lvq)    lvq_delete(lvq);       \
  if (table)  tab_delete(table,  0); \
  if (attmap) am_delete(attmap,  0); \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  DIMID c;                      /* loop variable for ref. vectors */
  int   n, k, m;                /* character counters */
  char  b[64];                  /* buffer for number output */

  assert(set && twrite);        /* check the function arguments */
  if (res.col_vec) {            /* if to write a vector index */
    if (mode & AS_ATT) {        /* if to write the header */
      twr_puts(twrite, res.col_vec);  /* write vector column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.col_act)) {
        n = (int)strlen(res.col_vec);
        k = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt);
        res.cwd_vec = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* compute width of prediction column */
      if (res.col_act) {        /* if to write a activation column */
        twr_fldsep(twrite);     /* write a field separator and */
        twr_puts(twrite, res.col_act);      /* the column name */
        if ((mode & AS_ALIGN) && (mode & AS_WEIGHT)) {
          n = (int)strlen(res.col_act);
          k = res.dig_act +3;   /* compute width of activation column */
          res.cwd_act = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
          if (k > n) twr_pad(twrite, (size_t)(k-n));
        }                       /* pad with blanks if requested */
      } }
    else {                      /* if to write a normal record */
      n = twr_printf(twrite, "%"DIMID_FMT, res.vec);
      if (res.cwd_vec > n) twr_pad(twrite, (size_t)(res.cwd_vec-n));
      if (res.col_act) {        /* if to write a activation column */
        twr_fldsep(twrite);     /* write separator and activation */
        n = twr_printf(twrite, "%.*g", res.dig_act,
                       lvq_activ(lvq, res.vec));
        if (res.cwd_act > n) twr_pad(twrite, (size_t)(res.cwd_act-n));
      }                         /* if to align, pad with blanks */
    } }
  else {                        /* if to write activations */
    if (mode & AS_ATT) {        /* if to write the header */
      if (!res.col_act) res.col_act = "";
      m = (int)strlen(res.col_act);
      if (mode & AS_ALIGN) {    /* if to align the column */
        n = snprintf(b, sizeof(b), "%"DIMID_FMT, res.cnt) +m;
        k = res.dig_act +3;     /* compute width of activ. columns */
        res.cwd_act = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      }
      for (c = 0; c < res.cnt; c++) {
        if (c > 0) twr_fldsep(twrite);
        twr_puts(twrite, res.col_act);
        n = m +twr_printf(twrite, "%"DIMID_FMT, c+1);
        if (res.cwd_act > n) twr_pad(twrite, (size_t)(res.cwd_act-n));
      } }                       /* print a column for each vector */
    else {                      /* if to write a normal record */
      for (c = 0; c < res.cnt; c++) {  /* traverse the vectors */
        if (c > 0) twr_fldsep(twrite); /* write a separator */
        n = twr_printf(twrite, "%.*g", res.dig_act, lvq_activ(lvq,c));
        if (res.cwd_act > n) twr_pad(twrite, (size_t)(res.cwd_act-n));
      }                         /* for each vector */
    }                           /* print the activation and */
  }                             /* pad the column for alignment */
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_lvq  = NULL;      /* name of neural network file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_inp  = NULL;      /* name of table file */
  CCHAR   *fn_out  = NULL;      /* name of output table file */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *comment = NULL;      /* comment characters */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_MARKED; /* table file read  mode */
  int     mout     = AS_ATT;           /* table file write mode */
  DIMID   dim      = -1;        /* data point/pattern dimension */
  DIMID   refcnt;               /* number of reference vectors */
  ATT     *att;                 /* to traverse the attributes */
  double  *vec;                 /* to traverse the patterns */
  ATTID   m, a, c;              /* number of attributes */
  DIMID   n, r;                 /* number of data points/vectors */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] lvqfile "
                    "[-d|-h hdrfile] tabfile outfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c#      name of the vector index field         "
                    "(default: activations)\n");
    printf("-m#      name of the activation field(s)        "
                    "(default: \"%s\")\n", res.col_act);
    printf("-o#      digits of neuron activations           "
                    "(default: %d)\n",     res.dig_act);
    printf("-a       align fields in output file            "
                    "(default: single separator)\n");
    printf("-w       do not write field names to output file\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("lvqfile  file to read reference vectors from\n");
    printf("-d       use default header "
                    "(field names = attribute numbers)\n");
    printf("-h       read table header "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write extended tuples to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': optarg = &res.col_vec;               break;
          case 'm': optarg = &res.col_act;               break;
          case 'o': res.dig_act = (int)strtol(s, &s, 0); break;
          case 'a': mout  |=  AS_ALIGN;                  break;
          case 'w': mout  &= ~AS_ATT;                    break;
          case 'r': optarg = &recseps;                   break;
          case 'f': optarg = &fldseps;                   break;
          case 'b': optarg = &blanks;                    break;
          case 'C': optarg = &comment;                   break;
          case 'n': mode  |=  AS_WEIGHT;                 break;
          case 'd': mode  |=  AS_DFLT;                   break;
          case 'h': optarg = &fn_hdr;                    break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_lvq = s;      break;
        case  1: fn_inp = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument and */
  if (k != 3) error(E_ARGCNT);  /* the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_lvq || !*fn_lvq) i++;
  if  (!fn_inp || !*fn_inp) i++;
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  mout |= AS_MARKED|AS_INFO1|AS_RDORD;
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse network/reference vectors --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) error(E_NOMEM);    /* for the reference vector set */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_lvq) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  matinp = (scn_first(scan) == T_ID)
        && (strcmp(scn_value(scan), "dom") != 0);
  if (matinp) {                 /* if matrix version */
    lvq = lvq_parse(scan, 0, 1);/* parse the input network */
    if (!lvq || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    m = lvq_dim(lvq);           /* get the number of inputs */
    fprintf(stderr, "[%"DIMID_FMT" input(s),", m); }
  else {                        /* if table version */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);      /* create an attribute set */
    if (as_parse(attset, scan, AT_ALL, 1) != 0)
      error(E_PARSE, scn_name(scan)); /* parse domain descriptions */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    a = as_attcnt(attset);      /* get the number of attributes */
    c = as_target(attset, NULL, -1);
    if (c < a)                  /* if there is a target attribute, */
      am_target(attmap, c);     /* set it in the attribute map */
    m = am_inpcnt(attmap);      /* get the number of inputs */
    lvq = lvq_parse(scan, m, 1);/* parse the neural network */
    if (!lvq || !scn_eof(scan, 1))
      error(E_PARSE, scn_name(scan));
    fprintf(stderr, "[%"DIMID_FMT" attribute(s),", a);
  }                             /* print a log message */
  refcnt = lvq_cnt(lvq);        /* get number of attributes/vectors */
  fprintf(stderr, " %"DIMID_FMT" vector(s)]", refcnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  res.cnt = refcnt;             /* note the number of vectors */

  if (matinp) {                 /* if matrix version */
    /* --- process patterns --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* open the output file */
    twr_xchars(twrite, recseps, fldseps, blanks, "");
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    k = vec_readx(&vec, &dim, tread);
    if (k) error(k, TRD_INFO(tread));
    if (dim < m) {              /* check the pattern size */
      free(vec); error(E_VECSIZE, dim); }
    for (r = 0; k == 0; r++) {  /* pattern read loop */
      for (c = 0; c < dim; c++) {
        twr_printf(twrite, "%.*g", res.dig_act, vec[c]);
        twr_fldsep(twrite);     /* print the pattern elements */
      }                         /* followed by a field separator */
      c = lvq_exec(lvq, vec, NULL);
      if (res.col_vec) {        /* if to print a vector index */
        twr_printf(twrite, "%"DIMID_FMT, c);  /* execute vectors */
        if (res.col_act) {      /* if to print membership degree */
          twr_fldsep(twrite);   /* print a separator */
          twr_printf(twrite, "%.*g", res.dig_act, lvq_activ(lvq, c));
        } }                     /* print the membership degree */
      else {                    /* if to print all m.s. degrees */
        for (c = 0; c < refcnt; c++) {
          if (c > 0) twr_fldsep(twrite);
          twr_printf(twrite, "%.*g", res.dig_act, lvq_activ(lvq, c));
        }                       /* print the membership degrees */
      }                         /* to the different vectors */
      twr_recsep(twrite);       /* and terminate the record */
      k = vec_read(vec, dim, tread);
    }                           /* read the next pattern */
    free(vec);                  /* delete the pattern buffer */
    if (k < 0) error(k, TRD_INFO(tread));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
    fprintf(stderr, "[%"DIMID_FMT" pattern(s)]", r);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t)); }

  else {                        /* if table version */
    /* --- mark input attributes --- */
    for (c = a; --c >= 0; ) {   /* traverse the parsed attributes */
      att = as_att(attset, c);  /* and mark the input attributes */
      att_setmark(att, (att_getdir(att) == DIR_IN) ? 1 : 0);
    }                           /* read only input attributes */

    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process table body --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_inp) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    if (mout & AS_ALIGN) {      /* if to align the output columns */
      table = tab_create("table", attset, tpl_delete);
      if (!table) error(E_NOMEM);  /* read the data table */
      k = tab_read(table, tread, mode);
      if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
      a = tab_attcnt(table);    /* get the number of attributes */
      n = tab_tplcnt(table);    /* and the number of tuples */
      w = tab_tplwgt(table);    /* and print a success message */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
      fprintf(stderr, " %"TPLID_FMT, n);
      if (w != (double)n) fprintf(stderr, "/%g", w);
      fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
      twrite = twr_create();       /* create a table writer and */
      if (!twrite) error(E_NOMEM); /* configure the characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      fprintf(stderr, "writing %s ... ", twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
      m += (res.col_vec ? (res.col_act ? 2 : 1) : res.cnt);
      vec = (double*)malloc((size_t)m *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      for (r = 0; r < n; r++) { /* traverse the tuples */
        tpl_toas(tab_tpl(table, r));     /* store in attribute set */
        am_exec(attmap, AM_INPUTS, vec); /* and convert to vector */
        res.vec = lvq_exec(lvq, vec, NULL);
        if (as_write(attset, twrite, mout, infout) != 0) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
      }                         /* write the current tuple */
      free(vec); }              /* delete the vector buffer */
    else {                      /* if to process tuples directly */
      k = as_read(attset, tread, mode);     /* read table header */
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      twrite = twr_create();       /* create a table writer and */
      if (!twrite) error(E_NOMEM); /* configure the characters */
      twr_xchars(twrite, recseps, fldseps, blanks, "");
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
      i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
      if (!(i & AS_DFLT))       /* if not done yet, read first tuple */
        k = as_read(attset, tread, mode);
      vec = (double*)malloc((size_t)m *sizeof(double));
      if (!vec) error(E_NOMEM); /* create a data point/vector buffer */
      for (w = 0, n = 0; k == 0; n++) {
        w += as_getwgt(attset); /* sum the tuple weight */
        am_exec(attmap, AM_INPUTS, vec);    /* convert to vector and */
        res.vec = lvq_exec(lvq, vec, NULL); /* execute the network */
        if (as_write(attset, twrite, mout, infout) != 0) {
          free(vec); error(E_FWRITE, twr_name(twrite)); }
        k = as_read(attset, tread, mode);
      }                         /* try to read the next tuple */
      free(vec);                /* delete the vector buffer */
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_delete(tread, 1);     /* delete the table reader */
      tread = NULL;             /* and clear the variable */
      m = as_attcnt(attset);    /* get the number of attributes */
    }
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  }                             /* if (matinp) .. else .. */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
